use warnings;
use strict;

package Geo::EOP;
use base 'XML::Compile::Cache';

use Geo::EOP::Util;   # all

use Log::Report 'geo-eop', syntax => 'SHORT';
use XML::Compile::Util  qw/unpack_type pack_type/;
use Math::Trig          qw/rad2deg deg2rad/;

# map namespace always to the newest implementation of the protocol
my %ns2version =
  ( &NS_HMA_ESA => '1.0'
  , &NS_EOP_ESA => '1.1'
  );

# list all available versions
my %info =
  ( '1.0'     =>
      { prefixes => {hma => NS_HMA_ESA, ohr => NS_OHR_ESA}
      , schemas  => [ 'hma1.0/*.xsd' ] }

  , '1.1'     =>
      { prefixes => {eop => NS_EOP_ESA, opt => NS_OPT_ESA}
      , schemas  => [ 'eop1.1/*.xsd' ] }

  , '1.2beta' =>
      { prefixes => {eop => NS_EOP_ESA, opt => NS_OPT_ESA}
      , schemas  => [ 'eop1.2beta/*.xsd'
                    , 'eop1.1/gmlSubset.xsd'
                    ] }
  );

my %measure =
  ( rad_deg   => sub { rad2deg $_[0] }
  , deg_rad   => sub { deg2rad $_[0] }
  , '%_float' => sub { $_[0] / 100 }
  , 'float_%' => sub { sprintf "%.2f", $_[0] / 100 }
  );
sub _convert_measure($@);

# This list must be extended, but I do not know what people need.
my @declare_always = ();

=chapter NAME
Geo::EOP - Earth Observation Products EOP (formely HMA)

=chapter SYNOPSIS
 use Geo::EOP ':eop1.1';

 my $eop = Geo::EOP->new('READER', version => '1.1');

 # see XML::Compile::Cache on how to use readers and writers
 my $data = $eop->reader("gml:GridCoverage")->($xmlmsg);
 my $xml  = $eop->writer($sometype)->($doc, $perldata);

 # or without help of the cache, XML::Compile::Schema
 my $r    = $gml->compile(READER => $sometype);
 my $data = $r->($xml);

 # huge overview on all defined elements
 $gml->printIndex;

=chapter DESCRIPTION
Provides access to the EOP definitions specified in XML, which is
based on GML.  Up to version 1.0, these schemas where named 'HMA'
(Heterogeneous EO Missions Accessibility), and the development is
still part of these ESA efforts for standardization.

BE WARNED that HMA1.0 and EOP 1.1/1.2 are only using a B<subset> of GML
version 3.1.1, due to problems with some software tools.  This limitation
will probably be removed in 1.3 (or whatever the next version will
be named).  EOP version 1.2-beta as defined here, is expected to become
the final 1.2 in December 2008, with only minor modifications.

The first releases of this module will not powerful, but hopefully
people contribute.  For instance, an example conversion script between
various versions is very welcome!  It would be nice to help each other.
I will clean-up the implementation, to make it publishable, but do not
have the knowledge about what is needed.

=chapter METHODS

=section Constructors
=c_method new 'READER'|'WRITER'|'RW', OPTIONS

=requires version VERSION|NAMESPACE
Only used when the object is created directly from this base-class.  It
determines which EOP syntax is to be used.  Can be a VERSION like "1.1"
or a NAMESPACE URI like 'NS_EOP_11'.

=option  prefixes ARRAY|HASH
=default prefixes undef
Prefix abbreviations, to be used by cache object.  Which prefixes are
defined depends on the schema version.

=option  allow_undeclared BOOLEAN
=default allow_undeclared <true>
In the optimal case, all types used in your application are declared
during the initiation phase of your program.  This will make it easy
to write a fast daemon application, or transform your program into a
daemon later.  So: "false" would be a good setting.  However, on the moment,
the developer of this module has no idea which types people will use.
Please help me with the specs!

=option  units HASH
=default units <none>
See L</Units>.  Only if you specify your preference in units of measurement,
some automatic translations will be performed.
=cut

sub new($@)
{   my ($class, $dir) = (shift, shift);
    $class->SUPER::new(direction => $dir, @_);
}

sub init($)
{   my ($self, $args) = @_;
    $args->{allow_undeclared} = 1
        unless exists $args->{allow_undeclared};

    $self->SUPER::init($args);

    $self->{GE_dir} = $args->{direction} or panic "no direction";

    my $version     =  $args->{version}
        or error __x"EOP object requires an explicit version";

    unless(exists $info{$version})
    {   exists $ns2version{$version}
            or error __x"EOP version {v} not recognized", v => $version;
        $version = $ns2version{$version};
    }
    $self->{GE_version} = $version;    
    my $info    = $info{$version};

    $self->prefixes
      ( xlink => NS_XLINK_1999, gml => NS_GML_311
      , sar => NS_SAR_ESA, atm => NS_ATM_ESA
      , %{$info->{prefixes}}
      );
    $self->addKeyRewrite('PREFIXED(sar,atm,ohr,opt)');

    (my $xsd = __FILE__) =~ s!\.pm!/xsd!;
    my @xsds    = map {glob "$xsd/$_"}
        @{$info->{schemas}}, 'xlink1.0.0/*.xsd';

    $self->importDefinitions(\@xsds);

    my $units = delete $args->{units};
    if($units)
    {   if($units->{angle})
        {   $self->addHook(type => pack_type(NS_GML_311, 'AngleType')
               , after => sub { _convert_measure $units->{angle}, @_} );
        }
        if($units->{distance})
        {   $self->addHook(type => pack_type(NS_GML_311, 'MeasureType')
               , after => sub { _convert_measure $units->{distance}, @_} );
        }
        if($units->{percentage})
        {   $self->addHook(path => qr/Percentage/
               , after => sub { _convert_measure $units->{percentage}, @_} );
        }
    }

    $self;
}

sub declare(@)
{   my $self = shift;

    my $direction = $self->direction;

    $self->declare($direction, $_)
        for @_, @declare_always;

    $self;
}

#---------------------------------

=section Accessors

=method version
EOP/HMA version, for instance '1.0'.

=method direction
Returns 'READER', 'WRITER', or 'RW'.
=cut

sub version()   {shift->{GE_version}}
sub direction() {shift->{GE_dir}}

=method printIndex [FILEHANDLE], OPTIONS
List all the elements which can be produced with the schema.  By default,
this only shows the elements and excludes the abstract elements from
the list.  The selected FILEHANDLE is the default to print to.
=cut

sub printIndex(@)
{   my $self = shift;
    my $fh   = @_ % 2 ? shift : select;
    $self->SUPER::printIndex($fh
      , kinds => 'element', list_abstract => 0, @_); 
}

#---------------------
# This code will probaby move to Geo::GML

sub _convert_measure($@)   # not $$$$ for right context
{   my ($to, $node, $data, $path) = @_;
    ref $data eq 'HASH'  or return $data;
    my ($val, $from) = @$data{'_', 'uom'};
    defined $val && $from or return $data;

    return $val if $from eq $to;
    my $code = $measure{$from.'_'.$to} or return $data;
    $code->($val);
}

#----------------------

=chapter DETAILS

=section Units
GML is careful about its units, which is nice but verbose.
Therefore, this module gives you a way to set your preferences.
The reader will simplify the structure, re-calculating the values
found into the units of measure you prefer, according to UCUM
(F<http://aurora.regenstrief.org/ucum>)

  my $eop = Geo::EOP->new(...
     , units => { distance => 'm'
                , angle    => 'deg' }
     );

Available conversions:
=over 4
=item angle 'deg'|'rad'
Will be applied to all AngleType elements
=item distance 'm'
Will be applied to all MeasureType elements
=item percentage '%'|'float'
Will be applied to all elements with a name ending on "Percentage".
I am not sure about the C<float>... cannot find the uom name.
=back

=cut

1;
